/*
 * $SunId: @(#)mdinfo.h 1.2 90/11/27 SMI [RMTC] $
 * Copyright (c) 1990 by Sun Microsystems, Inc.
 */
#ifndef __sundev_mdconfig_h
#define	__sundev_mdconfig_h

/*
 * Information structures for the mdgetconfig() library call.
 * Each node in the metadevice configuration tree is represented by
 * a common node type, struct md_node_info.
 * Most nodes contain pointers to node type specific information
 * structures.
 */
#include <sys/types.h>

/*
 * Common node information structure.
 * This is the structure that appears at each node in the
 * metadevice configuration tree.
 */
enum mdni_types	{
	MDNI_METAMIRROR,
	MDNI_METADEVICE,
	MDNI_STRIPE,
	MDNI_COMPONENT
};

struct md_node_info {
	enum mdni_types mdni_type;		/* Node type */
	union {					/* Type specific info structs */
		struct md_ni_metamirror *mdni_info_mm;
		struct md_ni_metadevice *mdni_info_md;
		struct md_ni_stripe	*mdni_info_st;
		struct md_ni_comp	*mdni_info_co;
		caddr_t			mdni_info_generic;
	} mdni_un;
	struct md_node_info *mdni_sib;		/* Pointer to sibling nodes */
	struct md_node_info *mdni_child;	/* Pointer to child nodes */
};

#define	mdni_info_metamirror	mdni_un.mdni_info_mm
#define	mdni_info_metadevice	mdni_un.mdni_info_md
#define	mdni_info_stripe	mdni_un.mdni_info_st
#define	mdni_info_component	mdni_un.mdni_info_co
#define	mdni_info		mdni_un.mdni_info_generic

/*
 * Metamirror node information structure.
 */
enum nimm_read_options {
	NIMM_RO_LOAD_BAL,
	NIMM_RO_GEOMETRY,
	NIMM_RO_FIRST
};

enum nimm_write_options {
	NIMM_WO_PARALLEL,
	NIMM_WO_SERIAL
};

struct md_ni_metamirror {
	dev_t nimm_dev;			/* unit # of metamirror */
	int nimm_submirrors;		/* Number of submirrors */
	enum nimm_read_options nimm_read_option;
	enum nimm_write_options nimm_write_option;
	int nimm_err_threshold;		/* Threshold error count */
	u_long nimm_total_blocks;	/* total disk blocks in metapartition */
	u_long nimm_rrd_blksize;	/* The blocksize of the dirty bits */
	u_long nimm_rrd_num;		/* The number of resync regions */
};

/*
 * Metadevice node information structure.
 * Metadevice nodes may be either simple metadevices or submirrors.
 * If a metadevice node is a submirror, it will point to a submirror
 * information structure.
 */
enum nimd_types {
	NIMD_SIMPLE,
	NIMD_SUBMIRROR
};

struct md_ni_metadevice {
	enum nimd_types nimd_type;	/* Submirror or not */
	int nimd_nrows;			/* Number of rows (stripes) */
	int nimd_hsp_id;		/* Hot spare pool id */
	struct md_ni_submirror *nimd_info; /* Submirror info */
};

/* Submirror information structure.
 * This is attached to a metadevice node if the metadevice is a
 * submirror.
 */
enum nism_submirror_states {
	NISM_UNATTACHED,	NISM_RUNNING,
	NISM_BOOT_RESYNC,	NISM_COMP_ERRED,
	NISM_COMP_RESYNC,	NISM_ATTACHED,
	NISM_ATTACHED_RESYNC,	NISM_OFFLINE,
	NISM_OFFLINE_RESYNC,	NISM_UNKNOWN
};

struct md_ni_submirror {
	short nism_unit;
	enum nism_submirror_states nism_state;
};

/*
 * Stripe node information structure.
 */
struct md_ni_stripe {
	int nist_ncomp;		/* Number of components in the stripe */
	int nist_blocks;	/* total blocks in this row */
	int nist_cum_blocks;	/* cumulative blocks in this row so far */
	int nist_ilsize;	/* stripe interleave size */
};

/*
 * Component node information structures.
 * All leaves in the tree should be this structure.
 */
enum nico_component_states {
	NICO_OKAY,
	NICO_ERRED,
	NICO_REPLACED,
	NICO_RESYNC,
	NICO_UNKNOWN
};

struct md_ni_comp {
	dev_t nico_dev;			/* dev_t of the physical partition */
	daddr_t nico_start_block;	/* component start blkno */
	enum nico_component_states nico_state; /* component state */
	int nico_replaced;		/* Component has been HS replaced */
	dev_t nico_hs_dev;		/* HS dev_t if HS replaced */
	short nico_ecounts;		/* component error count */
};

/*
 * Function definitions
 */
#ifndef KERNEL
struct md_node_info *mdgetinfotree(), *fmdgetinfotree();
void mdfreeinfotree();
#endif /* !KERNEL */
#endif /* __sundev_mdconfig_h */
