/*	@(#)mdreg.h 1.18 90/04/09 SMI	*/
/*	$SunId: @(#)mdvar.h 1.33 91/04/20 SMI [RMTC] $	*/

/*
 * Copyright (c) 1989 by Sun Microsystems, Inc.
 */

#include	<sys/ioccom.h>
#include	<sys/buf.h>
#include	<sundev/md_mddb.h>

/*
 * Debugging and other defines.
 */


#define	MD_DEBUG		0
#define	MD_IDEBUG		0
#define	MD_ODEBUG		0
#define	MD_DB_DEBUG		0
#define	MD_RR_DEBUG		0

/*
 * This is the current maximum number of real disks per Virtual Disk.
 */
#define	MD_ADM_MINOR		255	/* the minor number for /dev/rmdd */
#define	MD_MAX_ROWS		16	/* max rows of disks */
#define	MD_MAX_DISKS		4	/* max real disks per row */
#define	MD_MIN_RR_SIZE		200	/* 2000 blocks, or 100k */
#define	MD_DEF_NUM_RR		1000	/* Default number of dirty regions */
#define	MD_MAX_NUM_RR		(4192*NBBY - sizeof (struct optim_resync))
#define	MD_MAX_RESYNC_RETRIES	1	/* Max resync read retries for a unit */
#define	MD_SECTORS_PER_BLOCK	0x10

#define	MD_DELAY		10	/* 10 seconds delay */
#define	MD_MDELAY		10	/* 10 sec. md_mirror_timeout delay */

#define	NMIRROR			8

#define	MD_BUF_MAGIC		0x0BADCAFE	/* magic # for md_buf's */
#define	BUF_LOW_WATER_MARK	100		/* Max bufs kept in our pool */


/* memory user types for mdbuf_alloc() */
enum	mem_user_type {
	MUT_SLEEP, MUT_USE_RESERVE, MUT_NOSLEEP
};

/* position of request with respects to Resync Copy Area (rca) */
enum	md_rca_position {
	MD_BEHIND_RCA, MD_INTERSECTS_RCA, MD_AHEAD_RCA
};

/* daemon commands used in mds_daemon_cmd field above */
enum	md_daemon_cmds	{
	MDC_NONCMD,
	MDC_BUMP_ECOUNT, MDC_START_REQ, MDC_REDO_REQ,
	MDC_REQ_DONE,
	MDC_CHG_SM_STATE, MDC_CHG_COMP_STATE,
	MDC_RR_UPDATE,
	MDC_IOC_RESET, MDC_IOC_ATT_SM, MDC_IOC_SNARF_SET_SM,
	MDC_IOC_ONLINE, MDC_IOC_OFFLINE,
	MDC_IOC_ATTACH, MDC_IOC_DETACH,
	MDC_IOC_REPLACE, MDC_IOC_ADD,
	MDC_IOC_SETHS, MDC_IOC_GETHS,
	MDC_IOC_GETHSP, MDC_IOC_FIX,
	MDC_CHG_HSP_REFCOUNT,
	MDC_SLAVE_HS_RESYNC,
	MDC_IOC_CHANGE,
	MDC_CALL_STRATEGY,
};

/*
 * This #define is used in internal buf structures only.
 */
#define	b_mdsave b_chain

/*
 * Struct in which we save some essential information from the buffer
 * while passing it down to the real disk driver.
 * This structure is also used to pass requests to the metadaemon (metad).
 */
struct md_save {
	struct md_save	*mds_nextd;	/* doubly-linked list for daemon req. */
	struct md_save	*mds_prevd;
	struct md_save	*mds_nextq;	/* doubly-linked list for writes */
	struct md_save	*mds_prevq;	/*	during mirror resync's	 */
	unsigned int	mds_flags;	/* general flags, defined below */
	unsigned int	mds_dev;	/* metadevice that erred */
	unsigned int	mds_real_dev;	/* real device, i/o request sent to */
	unsigned int	mds_bflags;	/* bp->b_flag */
	unsigned int	mds_blkno;	/* bp->b_blkno */
	unsigned int	mds_bcount;	/* bp->b_bcount */
	enum md_daemon_cmds mds_daemon_cmd; /* metad request command */
	int	mds_error;	/* error returned from daemon request */
	struct buf	*mds_bp;	/* adr. of original metarequest buf */
	struct buf	*mds_back;	/* save bp->av_back, points to mdsave */
	struct md_save	*mds_mdsave;	/* back pointer for raw request */
	int		mds_frags;	/* # of fragmented raw requests */
	int		mds_frag_chars;	/* cum. byte count of finished frags */
	unsigned int	mds_current_sm;	/* current submirror index */
	unsigned int	mds_original_sm; /* current submirror index */
	unsigned int	mds_usable_sm;	/* readable/writable submirror bitmap */
	unsigned int	mds_alternate_sm; /* alt. readable submirrors bitmap */
	unsigned int	mds_accessed_sm; /* accessed submirrors bitmap */
	unsigned int	mds_erred_sm;	/* erred submirrors bitmap */
	struct md_components *mds_component; /* adr of component structure */
	unsigned int	mds_cleaned;
	caddr_t		mds_addr;
	unsigned long	mds_bumps;
};


/*
 * The following defines redefine some md_save fields to pass parameters
 * to the metadaemon.
 */
#define	mds_unit	mds_frags	/* unit number of metadevice, minor # */
#define	mds_state	mds_frag_chars	/* New state to change submirror to */
#define	mds_prev_state	mds_erred_sm	/* previous state befor change */


/* Flags used in mds_flags above. */
#define	MD_B_ERROR	 0x0001	/* an error has occurred */
#define	MD_B_IO_PENDING	 0x0002	/* long request still being broken up */
#define	MD_B_RETURN_ERR	 0x0004	/* waiting for daemon request and err */
#define	MD_B_REQ_DONE	 0x0008	/* daemon request is done */
#define	MD_B_SLOW_MODE	 0x0010	/* Mdreal is serially breaking up requests */
#define	MD_B_SERIAL_WR	 0x0020	/* Serial Write option for metamirrors */
#define	MD_B_WR_AFT_RD	 0x0040	/* Write Data after read, during opt. resync */
#define	MD_B_AFT_WRITING 0x0080	/* Write after read is active */

/*
 * enum declarations for mdstruct
 */
enum	md_types		{
	MD_UNDEFINED, MD_METAMIRROR, MD_SUBMIRROR, MD_DEVICE
};

enum	mm_write_options	{ WR_PARALLEL, WR_SERIAL };

enum	mm_read_options		{ RD_LOAD_BAL, RD_GEOMETRY, RD_FIRST };

enum	submirror_states	{
	SMS_UNATTACHED, SMS_RUNNING, SMS_BOOT_RESYNC,
	SMS_COMP_ERRED, SMS_COMP_RESYNC,
	SMS_ATTACHED, SMS_ATTACHED_RESYNC,
	SMS_OFFLINE, SMS_OFFLINE_RESYNC
};

enum	component_states	{
	CS_OKAY, CS_ERRED, CS_REPLACED, CS_RESYNC, CS_LAST_ERRED
};

enum	hotspare_states		{
	HSS_UNUSED, HSS_AVAILABLE, HSS_RESERVED, HSS_BROKEN
};

enum	mdi_retry	{
	MDI_RETRY_NOTHING, MDI_RETRY_OPEN
};

enum comp_lec_state {
	CLS_COVERED, CLS_OVERLAPS_LE, CLS_NOT_COVERED
};

/*
 * The md_struct struct describes one minor device of the Virtual Disk.
 * Each minor device is a either a metapartitions, which can be striped
 * across several rows, each consisting of an array of real partitions,
 * or a metamirror, which can have one or two mirror metapartitions under it.
 *
 * Here is the function of each field:
 *
 * md_type		There are three types to a metadevice:
 *			metamirror, submirror or metadevice.
 *
 * md_self_id		This is the minor device number for the metadevice
 *			which this structure represents.
 *
 * md_read_option	Metamirrors have several read options, this is the
 *			place whree the read option is stored.
 *
 * md_write_option	Metamirrors have two write options, they are
 *			parallel or serial.
 *
 * md_flag		configuration parameters. (i.e. LABEL and HAS_STRIPE)
 *
 * md_submirrors[]	each entry in the md_submirrors array is a
 *			representation of the submirrors within a metamirror.
 *			It includes: unit (minor device number), state,
 *					# erred components within the submirror
 *					and last block read (for optimizing
 *					the load balance algoritm.
 *
 * md_total_blocks	total blocks in this meta-partition.
 *
 * md_un.md_rows	total number of rows for this meta-partition
 * md_un.md_submirrors	number of submirror under a metamirror.
 *
 * md_metamirror	unit number of the metamirror for a given submirror.
 *			This is only used for submirrors.
 *
 * md_err_threshold	For metamirrors this is the error threshold for reads.
 *
 * md_pass_num		no use to the kernel. it's here only for the
 *			convenience of crash recovery.
 *
 * For each row:
 *
 *	md_width:		Width of this row of the meta-partition	in
 *				real partitions.
 *
 *	md_start_block:		Starting block number in the real partition.
 *
 *	md_blocks:		Total number of blocks in this row.
 *				This is used in calculating the cumulative
 *				total of blocks at each row.
 *
 *	md_cum_blocks:		The cumulative number of blocks up to this row.
 *				This is used in calculating which row the
 *				current	block falls into.
 *
 *	md_real_disks[]:	List of real partition used within the
 *				metadevice. This structure includes the
 *				device number, start blkno, amount space
 *				acutally being used, real size of the physical
 *				device, # error counts on the device.
 *
 */
struct md_struct	{
	enum md_types		md_type;
	dev_t			md_self_id;
	mddb_recid_t		md_record_id;
	enum mm_read_options	md_read_option;
	enum mm_write_options	md_write_option;
	unsigned short		md_flag;	/*  configuration info */
	struct	md_submirrors {			/* submirrors */
		short	unit;
		enum submirror_states state;
	} md_submirrors[NMIRROR];
	daddr_t	md_total_blocks;	/* total disk blocks in metapartition */
	union {
		short	md_rows;	/* number of rows */
		short	md_submirrors;	/* number of submirrors */
	} md_un;
	short	md_metamirror;		/* unit # of metamirror, sm use only. */
	short	md_err_threshold;	/* # errors til offlining components */
	short	md_pass_num;		/* this is here for convience */
	dev_t	md_rdev;		/* associated raw device for ioctl */
	struct md_row	{
		short	md_width;	/* width of this row */
		int	md_blocks;	/* total blocks in this row */
		int	md_cum_blocks;	/* cum. blks in this and prev. rows */
		int	md_interlace;	/* no blks from each disk in a stripe */
		struct md_components {	/* components */
			dev_t	devnum;		/* device number */
			daddr_t	start_block;	/* component start blkno */
			enum component_states state; /* component state */
			short	ecounts;	/* component error count */
			int	hot_spared;	/* flag, specifying hs */
			dev_t	hot_spare;	/* hs dev number */
			daddr_t	hot_start_block; /* start block of hs */
		}md_components [MD_MAX_DISKS];
	} md_real_row[MD_MAX_ROWS];
	u_long	md_rrd_blksize;	/* The blocksize of the dirty bits */
	u_long	md_rrd_num;		/* The number of resync regions */
	mddb_recid_t	md_rr_dirty_recid;
	int	md_hsp_id;		/* hot spare pool id for metadevice */
};

struct mdstruct_incore {
	struct md_struct *mdi_struct;	/* pointer to main md_struct */
	unsigned short	mdi_status;
	unsigned long	mdi_breaking_io;   /* # of requests being broken up */
	unsigned long	mdi_outstanding;   /* # outstanding I/Os */
	unsigned long	mdi_bump_ecounts;  /* # of BUMP_ECOUNT requests */
	unsigned long	mdi_total_bumps;  /* total # of BUMP_ECOUNT requests */
	unsigned long	mdi_hs_need_resync; /* # hs needing to be resynced */
	struct buf	*mdi_rbuf;	   /* raw I/O buf structure */
	struct vnode	*mdi_specnodes[MD_MAX_ROWS][MD_MAX_DISKS]; /* vnodes */
	struct hot_spare *mdi_hs_struct[MD_MAX_ROWS][MD_MAX_DISKS];
	int		mdi_last_read;	   /* last submirror index read */
	struct md_incore_sync_struct *missp;	/* if null, no sync? */
	char		*mdi_rtbp;	/* pointer to resync trans bits */
	char		*mdi_ttdbp;	/* transiting to dirty bitmap */
	short		*mdi_owp;	/* outstanding write array */
	char		*mdi_rdbp;	/* Resync region dirty bits */
	enum mdi_retry	mdi_retry;	/* Retry enum, tells open what to do */
	char		*mdi_serial_bm;	/* serialize bm, strat. during resync */
	char		*mdi_orig_rdbp;	/* original resync region dirty bits */
	unsigned long	mdi_waiting_io;	/* # processes waiting on out. io */
};

/*
 * Flags per minor device which indicate current status. All used in mdi_status.
 * These flags are set/reset only by the kernel.
 */
#define	MD_SETUP	  0x0001 /* partition has been init'ed */
#define	MD_BOPEN	  0x0002 /* block metapartition has been opened */
#define	MD_COPEN	  0x0004 /* raw metapartition has been opened */
#define	MD_RESYNC_ACTIVE  0x0008 /* resync is in progress */
#define	MD_RESYNC_BREAK	  0x0010 /* stall point for resync process */
#define	MD_RESET_PENDING  0x0020 /* Disallow opens, metadevice is being reset */
#define	MD_WAITING_BREAK  0x0040 /* waiting for mdreal to finish */
#define	MD_WAITING_STATE  0x0080 /* waiting for state change to finish */
#define	MD_WAITING_RESYNC 0x0100 /* waiting for resync stall point */
#define	MD_MTIMEOUT	  0x0200 /* Metamirror is on md_mirror_timeout queue */
#define	MD_KEEP_DIRTY	  0x0400 /* Don't run the mirror dirty bit sm */
#define	MD_STALL_STRAT	  0x0800 /* Stall all new strategy calls */
#define	MD_STALL_RESYNC	  0x1000 /* Stall all md_mirror_copy at top of loop */
#define	MD_SWAP		  0x2000 /* device found on list of swap devices */
#define	MD_WAITING_RCA	  0x4000 /* requests are pended waiting on rca */
#define	MD_FREEZE	  0x8000 /* Stop all new strategy/ioctl - DEBUGGING */

/*
 * The following two flags are in md_flag field of md_struct.
 */
#define	MD_HAS_LEC	0x20	/* Metamirror has a last errored component */
#define	MD_LABEL_FIRST	0x40	/* First sector of the metadevice is a label */
#define	MD_HAS_STRIPE	0x80	/* metapartition has striping (md_width != 1) */
#define	MD_COMPAT	0x0100	/* MDD-1 compatible metadevice */

/*
 * The following are macros that are used to lock and unlock ioctl
 * accesses to the mdstructs
 */
#define	MD_ULOCKBIT	0x1

#define	MD_LOCK(unit)	{ \
	if ((unit) < 0 || (unit) > NMD) \
		panic("MD_LOCK"); \
	while (md_unit_locks[unit] & MD_ULOCKBIT) { \
		md_lock_sleeps++; \
		(void) sleep(&md_unit_locks[unit], PRIBIO); \
	} \
	md_unit_locks[unit] |= MD_ULOCKBIT; \
}

#define	MD_UNLOCK(unit) { \
	if (md_unit_locks[unit] == 0) \
		panic("MD_UNLOCK"); \
	md_unit_locks[unit] = 0; \
	wakeup(&md_unit_locks[unit]); \
}

#define	OR_MAGIC	0xFECA	/* Only missing the L */
/* Structure for optimized resync records */
struct optim_resync {
	u_long or_magic;
	u_long or_blksize;
	u_long or_num;
	char or_rr[1];
};


/* user allocatable stuff */
struct md_buf {
	struct md_save	mdb_save;
	struct buf	mdb_buf;
};

/* user allocatable stuff, plus header for mdbuf_alloc() and mdbuf_free() */
struct md_buf_and_hdr {
	struct md_save		mdbh_save;
	struct buf		mdbh_buf;
	struct md_buf_and_hdr	*mdb_head;
	u_int			mdb_nalloc;
	u_int			mdb_nchunk;
	int			mdb_refcnt;
	u_int			mdb_magic;
};

struct md_mmtq {	/* The metamirror timeout queue	 struct */
	struct md_mmtq *next;
	struct md_mmtq *prev;
	struct md_save sync;
	int valid;	/* Used to mark invalid mmtqs */
	int unit;	/* the unit name of the metamirror */
};

#define	MMTQ_NULL	((struct md_mmtq *)0)

enum sync_type { MS_UNKNOWN, MS_USER_SINGLE, MS_USER_ALL, MS_RESYNC_ALL,
	MS_PARTIAL, MS_OPTIMIZED, MS_USER_OPTIMIZED, MS_PARTIAL_ALL,
	MS_USER_PARTIAL_ALL, MS_FULL_OPTIM, MS_COMP_OPTIM };

enum sync_error { MSE_UNKNOWN, MSE_NONE_TO_SYNC, MSE_INVALID_TYPE,
	MSE_NOT_SETUP, MSE_NOT_METAMIRROR, MSE_1WAY_MIRROR, MSE_O_2READABLE,
	MSE_DB_ERROR, MSE_INVALID_SUBMIRROR, MSE_INVALID_COMP,
	MSE_NONSYNCABLE_SUBMIRROR, MSE_NONSYNCABLE_COMP, MSE_NO_SOURCES,
	MSE_INVALID_COPYSIZE, MSE_ALREADY_SYNCING, MSE_NO_ERR,
	MSE_CANT_TRANS_COMP, MSE_CANT_TRANS_SM, MSE_INTERRUPTED };


/*
 * Hot spare and hot spare pool data structures
 */
struct hot_spare {
	mddb_recid_t		hs_record_id;	/* db record id */
	struct hot_spare	*hs_next;	/* hs list, link */
	dev_t			hs_devnum;	/* hs device number */
	daddr_t			hs_start_blk;	/* hs starting block */
	int			hs_has_label;	/* hs has a label */
	u_long			hs_number_blks;	/* hs # of blocks */
	enum hotspare_states	hs_state;	/* hs state */
	int			hs_refcount;	/* # hsp using the hs */
	struct vnode		*hs_specnode;	/* ptr to vnode for hs */
};

struct hot_spare_pool {
	int			hsp_self_id;	/* hsp identifier */
	mddb_recid_t		hsp_record_id;	/* db record id */
	struct hot_spare_pool	*hsp_next;	/* hsp list, link */
	int			hsp_refcount;	/* # metadevices using hsp */
	int			hsp_nhotspares;	/* # hs in the pool */
	dev_t			hsp_hs_devs[1];	/* array of dev_t per hs */
};



/* Global flags */
#define	MD_DAEMON_RUNNING	0x0001	/* master daemon has been started. */
#define	MD_K_DAEMON_RUNNING	0x0002	/* kernel daemon has been started. */
#define	MD_S_DAEMON_RUNNING	0x0004	/* slave daemon has been started. */
#define	MD_DB_SNARFED		0x0008	/* incore structs built for db recs */
#define	MD_DAEMON_GO_AWAY	0x0010	/* Bit to kill multiple daemons. */
#define	MD_DB_STALE		0x0020	/* data base not known to be correct */
#define	MD_HS_RESYNC_NEEDED	0x0040	/* hot spare resync needed */
#define	MD_SLAVE_SCANNING	0x0080	/* slaved scanning hs to be resync */
#define	MD_SHUTDOWN		0x0100	/* driver is being totally shutdown */
#define	MD_DB_SNARFING		0x0200	/* driver is in process of snarfing */
#define	MD_MASTERD_LOCKED	0x0400	/* master daemon is locked */
#define	MD_MASTERD_LOCK_WANTED	0x0800	/* master daemon lock is wanted */
/*
 * At a later stage, we may decide to have extra bits in the
 * dev_t, and then we should mask them out in this macro.
 */
#define	MD_MINOR(dev)	minor(dev)

/*
 * ioctl codes.
 */
#define	MD_IOC_OLDGET	_IOR(V, 0, struct md_old_struct) /* get old config. */
	/* set configuration */
#define	MD_IOCSET	_IOWR(V, 1, struct md_set_params)
#define	MD_IOCMSYNC	_IOWR(V, 2, struct md_sync_struct) /* mirror resync */
#define	MD_IOCDAEMON	_IOWR(V, 3, struct daemon_params) /* create md daemon */
#define	MD_IOCRESET	_IOWR(V, 4, struct md_i_reset)	/* clear config. */
	/* offline a submirror */
#define	MD_IOCOFFLINE	_IOWR(V, 5, struct md_i_offline)
#define	MD_IOCONLINE	_IOWR(V, 6, struct md_i_online)	/* online submirror */
				/* attach a submirror */
#define	MD_IOCATTACH	_IOWR(V, 7, struct md_att_struct)    /* attach a sm */

#define	MD_IOCDETACH	_IOWR(V, 8, struct md_detach_params) /* detach a sm */
#define	MD_IOCGET	_IOWR(V, 9, struct md_i_get)	/* get configuration */
	/* replace component */
#define	MD_IOCREPLACE	_IOWR(V, 10, struct replace_params)
#define	MD_IOCADD	_IOWR(V, 11, struct md_add_struct)  /* add devices */
#define	MD_IOCNUM	_IOR(V, 12, int)    /* Get supported # of metadevices */
#define	MD_IOCSET_HS	_IOWR(V, 13, struct set_hs_params) /* setup h.s. */
#define	MD_IOCGET_HS	_IOWR(V, 14, struct get_hs_params) /* get h.s. */
#define	MD_IOCGET_HSP	_IOWR(V, 15, struct get_hsp_params) /* get h.s.p. */
#define	MD_IOCFIX	_IOWR(V, 16, struct fix_params)	/* fix components */
	/* Get the size of the resync region map */
#define	MD_IOCRRMSIZE	_IOWR(V, 17, struct md_rrsize)
#define	MD_IOCGETRR	_IOWR(V, 18, struct md_rrget) /* Get resync region bm */
	/* Get sync status */
#define	MD_IOCGETSS	_IOWR(V, 19, struct md_incore_sync_struct)
#define	MD_HALT		_IO(V, 20) /* driver state as if never accessed */
#define	MD_IOCCHANGE	_IOWR(V, 21, struct md_change_params) /* chg params */

/* Debugging ioctls */

#define	MD_IOCGETRRTD	_IOW(V, 200, struct md_rrtdget) /* get trans data */
#define	MD_IOCFNC	_IOWR(V, 201, struct md_fnc)	/* Freeze and compare */

/*
 * ioctl parameter structures
 */

/* Generic error structure for metadevice ioctls */

enum md_errors {
	MDE_NONE, MDE_INVAL_UNIT, MDE_SNARF_FAILURE, MDE_UNIT_NOT_SETUP,
	MDE_UNIT_ALREADY_SETUP,
	MDE_SM_NOT_SETUP, MDE_INVAL_SM_UNIT,
	MDE_SM_IN_USE, MDE_IS_OPEN, MDE_RESYNC_ACTIVE, MDE_SM_IS_OPEN,
	MDE_IS_SM, MDE_SM_TOO_SMALL, MDE_NO_LABELED_SM,
	MDE_SM_IS_METAMIRROR, MDE_MMN_SMN_MISMATCH, MDE_INVAL_ROWS,
	MDE_INVAL_WIDTH, MDE_INTERLACE_ERROR, MDE_INVAL_COMP,
	MDE_COMP_SIZE, MDE_TB_ERROR,
	MDE_LOCK_FAILURE, MDE_COMMIT_FAILURE,
	MDE_HSP_CREATE_FAILURE, MDE_HS_CREATE_FAILURE, MDE_CREATE_FAILURE,
	MDE_LOCKS_FAILURE, MDE_COMMITS_FAILURE,
	MDE_DAEMON_RUNNING, MDE_KDAEMON_RUNNING,
	MDE_NSLAVED,
	MDE_NOT_SM, MDE_NOT_MM,
	MDE_MM_NOT_MM, MDE_SYNC_INTERRUPTED,
	MDE_CANT_FIND_SM, MDE_NO_RUNNING_SM, MDE_ILLEGAL_SM_STATE,
	MDE_MIRROR_FULL, MDE_LAST_SM, MDE_SM_FAILED_COMPS,
	MDE_LAST_SM_RE, MDE_CANT_FIND_COMP, MDE_COMP_TOO_SMALL,
	MDE_COMP_OPEN_ERR, MDE_HS_OPEN_ERR,
	MDE_ROW_FULL, MDE_RR_ALLOC_ERROR,
	MDE_INVAL_HSOP, MDE_HS_INUSE, MDE_INVAL_HS, MDE_INVAL_HSP,
	MDE_HSP_BUSY, MDE_HSP_REF, MDE_HS_RESVD, MDE_HS_COPYOUT_ERR,
	MDE_NO_HSP,
	MDE_GHSP_SIZE_ERROR, MDE_FIX_INVAL_STATE, MDE_FIX_SIZE_ERROR,
	MDE_FIX_INVAL_HSP, MDE_FIX_INVAL_HS_STATE, MDE_RRTD_SIZE_ERROR,
	MDE_RR_SIZE_ERROR, MDE_RTBP_COPYOUT_ERR, MDE_OWP_COPYOUT_ERR,
	MDE_NO_RESYNC_ACTIVE, MDE_DB_STALE, MDE_DB_NOT_FOUND,
	MDE_IS_STILLREF, MDE_C_WITH_INVAL_SM,
	MDE_FIX_RESYNC_ACTIVE, MDE_REPL_RESYNC_ACTIVE,
	MDE_DATABASE_FULL, MDE_NOT_ENOUGH_DBS,
	MDE_REPL_INVAL_STATE, MDE_NO_READABLE_SM,
	MDE_INVAL_MM_PARAM, MDE_INVAL_MD_PARAM, MDE_HS_IN_USE };

struct md_error {
	enum	md_errors error;
	int	unit;
	int	sm_unit;
	int	dev;
};


struct md_set_params {
	struct md_error		mde;
	int			unit;
	int			metamirror;
	struct md_struct	*mdstruct;
	int			size;
};

/* Can only replace components */
struct replace_params {
	struct md_error	mde;
	enum sync_error sync_error;		/* Sync error  */
	dev_t		old_device;	/* enable/replace use this */
	dev_t		new_device;	/* replace only uses this */
	int		do_sync;	/* 0 - no sync, <> 0, sync */
	int		enable_comp;	/* 0 - replace comp, 1 - enable comp */
	daddr_t		start_blk;	/* start block of new device */
	int		has_label;	/* has label flag of new device */
	u_long		number_blks;	/* # of blocks of new device */
};

struct md_i_reset {
	struct md_error		mde;
	int			dev;		/* Unit to clear */
	int			recursive;	/* Do it recursively */
	int			force;		/* Force the clear */
};

struct md_i_offline {
	struct md_error		mde;
};

struct md_i_online {
	struct md_error		mde;		/* Normal error */
	int			do_sync;	/* <> 0, do sync */
	enum sync_error sync_error;		/* Sync error  */
};

struct md_att_struct {
	struct md_error		mde;		/* Normal error */
	int			dev_to_attach;	/* The device  to attach */
	int			do_sync;	/* 0 don't sync, <>0 sync */
	enum sync_error		sync_error;	/* sync error */
};

struct md_detach_params {
	struct md_error		mde;
	int			sm_unit;
	int			force_detach;
};

struct md_i_get {
	struct md_error		mde;
	int			unit;
	struct md_struct	*mdp;
};

struct md_add_struct {
	struct md_error	mde;		/* Error return */
	int		num_devs;	/* Number of valid entries */
	int		total_blocks;	/* Total # of blocks in the row */
	int		interlace;	/* Interlace value of stripe	*/
	struct {
		dev_t	dev_num;
		daddr_t start_blk;
	} add_devs[MD_MAX_DISKS];
};

struct md_rrsize {
	struct md_error	mde;		/* error return */
	int		unit;		/* unit # to get */
	u_long		rr_num;		/* Number of resync regions */
	u_long		rr_blksize;	/* Blocksize of regions */
};

struct md_rrget {
	struct md_error mde;
	int		unit;		/* unit # to get */
	u_long		array_size;
	char		*array;	/* pointer to array[howmany(size, NBBY)] */
};

struct md_rrtdget {
	struct md_error mde;
	u_long		array_size;
	char		*rtbp_array;	/* trans bits (num+NBBY)/NBBY bytes */
	short		*owp_array;	/* For owp shors (num shorts) */
};

struct md_fnc { /* Freeze and compare */
	struct md_error mde;
	u_long	action;		/* 1, freeze, 0 don't freeze */
	u_long	status;		/* See below */
	u_long	num_diffed;	/* Number of blocks that diffed */
	u_long	num_same;	/* Number of blocks that were the same */
#define	FNC_ERRORED	0x1
#define	FNC_FAILURE	0x2
#define	FNC_SUCCESS	0x3
	u_long	maxread;
	u_long	minread;
	u_long	nread;
};

struct md_sync_struct {
	struct md_error mde;
	enum sync_type	mss_sync_type;
	enum sync_error mss_sync_error;
	int		mss_copysize;	/* The size of the copy buffer */
	int		mss_to_unit;	/* Specifies unit for MS_USER_SINGLE */
	dev_t		mss_to_comp;	/* The component, for MS_PARTIAL */
	int		mss_current_copy; /* returned, in case of EINTR */
	int		mss_zerofill;	/* Zerofill on lec read error */
};

#define	MISS_RUNNING	1
#define	MISS_STOPPED	2
struct md_incore_sync_struct {
	struct md_error	mde;	/* Only for MD_IOCGETSS */
	int	unit;		/* Only for MD_IOCGETSS */
	int	miss_status;		/* Only for MD_IOCGETSS */
	int	miss_pid;		/* Only for MD_IOCGETSS */
	enum sync_type miss_sync_type;
	enum sync_error miss_sync_error;
	int miss_copysize;
	int miss_to_unit;	/* if MS_USER_SINGLE or MS_PARTIAL */
	int miss_to_index;	/* submirror index of the unit */
	dev_t miss_to_compdev;	/* if MS_PARTIAL */
	int   miss_zerofill;	/* zerofill on lec read error */
	daddr_t	miss_current_copy;  /* blkno of current mirror resync */
	unsigned int miss_usable_sm;	/* Running only (for reading) */
	unsigned int miss_alternate_sm;	/* Read from here when running fail */
	unsigned int miss_accessed_sm;	/* Submirrors that have been accessed */
	unsigned int miss_erred_sm;	/* Submirrors that have erred */
	int miss_dests;		/* destinations */
	int miss_resync_start;	/* Starting block */
	int miss_resync_end;	/* Ending block */
	int miss_resync_blksize;	/* blocksize (if MS_PARTIAL) */
	int miss_resync_skipsize;	/* skipsize (if MS_PARTIAL) */
	daddr_t miss_rca_blkno;	/* For synchronizing resync & strategy */
	int miss_rca_nblks;	/* For synchronizing resync & strategy */
};

enum set_hs_command		{
	ADD_HOT_SPARE, DELETE_HOT_SPARE, REPLACE_HOT_SPARE
};

struct set_hs_params {
	struct md_error		mde;			/* error return */
	enum set_hs_command	shs_cmd;		/* ioctl command */
	int			shs_hot_spare_pool;	/* hsp identifier */
	dev_t			shs_component_old; /* dev for add, del, repl */
	dev_t			shs_component_new;	/* new dev for repl */
	int			shs_options;
	daddr_t			shs_start_blk;		/* used by add/repl */
	int			shs_has_label;		/* used by add/repl */
	u_long			shs_number_blks;	/* used by add/repl */
};

struct get_hs_params {
	struct md_error		mde;			/* error return */
	dev_t			ghs_devnum;		/* hs device number */
	daddr_t			ghs_start_blk;		/* returned start blk */
	daddr_t			ghs_number_blks;	/* returned # of blks */
	enum hotspare_states	ghs_state;	/* returned state */
};


struct get_hsp_params {
	struct md_error	mde;		/* error return */
	int		ghsp_self_id;	/* hsp identifier */
	int		ghsp_next_id;	/* next hsp identifier */
	int		ghsp_nhotspares;	/* # of hs in hsp */
	int		ghsp_size;	/* size of piece of memory */
	dev_t		*ghsp_hs_devs;	/* ptr to array(user space) of devv's */
};


struct fix_params {
	struct md_error		mde;		/* Error return */
	enum sync_error		sync_error;	/* Sync error return */
	dev_t			fix_component;	/* dev_t of failed component */
};

struct md_change_params {
	struct md_error		mde;		/* Error return */
	int			change_read_option;
	enum mm_read_options	new_read_option;
	int			change_write_option;
	enum mm_write_options	new_write_option;
	int			change_err_threshold;
	short			new_err_threshold;
	int			change_pass_number;
	short			new_pass_number;
	int			change_hot_spare_pool;
	int			new_hsp_id;
};

struct daemon_params {
	struct md_error	mde;
	int		daemon_type;
	int		num_slave_daemons;
};

int	mddone();


#define	dprintf		if (md_debugging) printf
#define	dprintf1	if (md_debugging && md_debug > 0) printf
#define	dprintf2	if (md_debugging && md_debug > 1) printf
#define	dprintf3	if (md_debugging && md_debug > 2) printf

#define	drrprintf	if (md_debugging) printf
#define	drrprintf1	if (md_debugging && md_rr_debug > 0) printf
#define	drrprintf2	if (md_debugging && md_rr_debug > 1) printf
#define	drrprintf3	if (md_debugging && md_rr_debug > 2) printf



#define	md_bahhold(mdbp) \
	if ((mdbp)->mdb_refcnt == 0) \
		panic("hold on unreferenced savbuf"); \
	(mdbp)->mdb_refcnt++;

#define	md_bufhold(savp) { \
	int	ss; \
	ss = splbio(); \
	md_bahhold((struct md_buf_and_hdr*)(savp)); \
	(void) splx(ss); \
}
#define	md_savhold(savp)	md_bufhold((struct md_buf*)(savp));

/* insert/remove mdsave to/from md_daemon_queue */
#define	remd(mds) { \
	int	s; \
	if (((mds)->mds_prevd == NULL) || ((mds)->mds_nextd == NULL)) \
		panic("remd\n"); \
	s = splbio(); \
	(mds)->mds_prevd->mds_nextd = (mds)->mds_nextd; \
	(mds)->mds_nextd->mds_prevd = (mds)->mds_prevd; \
	(void) splx(s); \
}

#define	instaild(mds, mds1) { \
	int	pri; \
	if ((mds1)->mds_prevd == NULL) \
		panic("instaild\n"); \
	pri = splbio(); \
	md_savhold((mds)); \
	(mds1)->mds_prevd->mds_nextd = (mds); \
	(mds)->mds_prevd = (mds1)->mds_prevd; \
	(mds1)->mds_prevd = (mds); \
	(mds)->mds_nextd = (mds1); \
	(void) splx(pri); \
}

#define	send_daemon_requestxx(mds, cmd, queue) { \
	int	pri; \
	(mds)->mds_daemon_cmd = (cmd); \
	if ((queue)->mds_prevd == NULL) \
		panic("send_daemon_request\n"); \
	pri = splbio(); \
	md_savhold((mds)); \
	(queue)->mds_prevd->mds_nextd = (mds); \
	(mds)->mds_prevd = (queue)->mds_prevd; \
	(queue)->mds_prevd = (mds); \
	(mds)->mds_nextd = (queue); \
	(void) splx(pri); \
	wakeup((caddr_t)(queue)); \
}


#define	BLK_TO_RR(i, b, mp)  {\
	(i) = ((b) / ((mp)->md_rrd_blksize)); \
	if ((i) > (mp)->md_rrd_num) { panic("md: BLK_TO_RR"); } \
}

#define	RR_TO_BLK(b, i, mp) \
	(b) = ((i) * ((mp)->md_rrd_blksize))

#define	IS_RR_DIRTY(i, midp) (isset((midp)->mdi_rdbp, (i)))
#define	CLR_RR_DIRTY(i, midp) (clrbit((midp)->mdi_rdbp, (i)))
#define	SET_RR_DIRTY(i, midp) (setbit((midp)->mdi_rdbp, (i)))

#define	IS_RR_TRANS(i, midp) (isset((midp)->mdi_rtbp, (i)))
#define	CLR_RR_TRANS(i, midp) (clrbit((midp)->mdi_rtbp, (i)))
#define	SET_RR_TRANS(i, midp) (setbit((midp)->mdi_rtbp, (i)))

#define	IS_RR_TODIRTY(i, midp) (isset((midp)->mdi_ttdbp, (i)))
#define	CLR_RR_TODIRTY(i, midp) (clrbit((midp)->mdi_ttdbp, (i)))
#define	SET_RR_TODIRTY(i, midp) (setbit((midp)->mdi_ttdbp, (i)))


#define	SUBMIRROR_IS_WRITABLE(mdstruct, isubmirror) \
	((mdstruct)->md_submirrors[(isubmirror)].state == SMS_RUNNING || \
	(mdstruct)->md_submirrors[(isubmirror)].state == SMS_BOOT_RESYNC || \
	(mdstruct)->md_submirrors[(isubmirror)].state == SMS_COMP_ERRED || \
	(mdstruct)->md_submirrors[(isubmirror)].state == SMS_COMP_RESYNC || \
	(mdstruct)->md_submirrors[(isubmirror)].state == SMS_ATTACHED_RESYNC ||\
	(mdstruct)->md_submirrors[(isubmirror)].state == SMS_OFFLINE_RESYNC)

#define	SUBMIRROR_IS_READABLE(mdstruct, isubmirror) \
	((mdstruct)->md_submirrors[(isubmirror)].state == SMS_RUNNING || \
	(mdstruct)->md_submirrors[(isubmirror)].state == SMS_BOOT_RESYNC || \
	(mdstruct)->md_submirrors[(isubmirror)].state == SMS_COMP_ERRED || \
	(mdstruct)->md_submirrors[(isubmirror)].state == SMS_COMP_RESYNC)

#define	ERRED_COMPONENT(state) \
	(((state) == CS_ERRED) || \
	((state) == CS_REPLACED) || \
	((state) == CS_LAST_ERRED))


/*
 * ========================================================================
 * The definitions below should NEVER be modified!!!!!
 * These definitions are the mdd-1 (old) definitions.
 * This is all so we can create one ioctl for ioadm to use.
 */

/*
 * Configuration information from MDD-1
 */
#define	OLD_NMD			24
#define	MD_OLD_MAX_ROWS		4
#define	MD_OLD_MAX_DISKS	4
#define	OLD_NMIRROR		2
#define	MD_OLD_MAGIC		3922319
#define	MD_OLD_VERSION		1

/*
 * md_status values from MDD-1.
 */
#define	MD_OLD_SETUP		0x1	/* Device has been metainit'ed */
#define	MD_OLD_BOPEN		0x2	/* Block device has been opened */
#define	MD_OLD_COPEN		0x4	/* Raw device has been opened */
#define	MD_OLD_RESYNC_FTOS	0x10	/* Doing resync from first to second */
#define	MD_OLD_RESYNC_STOF	0x20	/* Doing resync from second to first */
#define	MD_OLD_MIRROR		0x40	/* This is a submirror */

/*
 * md_flag values from MDD-1.
 */
#define	MD_OLD_VOLUME		0x1	/* This is a metamirror */
#define	MD_OLD_OFFLINED		0x2	/* Metapartition has been offlined */
#define	MD_OLD_PENDING_FTOS	0x4	/* Pending resync from 1 to 2 */
#define	MD_OLD_PENDING_STOF	0x8	/* Pending resync from 2 to 1 */
#define	MD_OLD_SEQ_WRITE	0x10	/* Serial write mode enabled */
#define	MD_OLD_FIRST_ONLY	0x20	/* Read only from the first mirror */
#define	MD_OLD_LABEL_FIRST	0x40	/* First sector is a disk label */
#define	MD_OLD_HAS_STRIPE	0x80	/* (md_width != 1) */
#define	MD_OLD_GEOMETRY_RD	0x100	/* Geometric read option enabled */

struct	md_old_save {
	struct md_old_save	*mds_nextd;
	struct md_old_save	*mds_prevd;
	struct md_old_save	*mds_nextq;
	struct md_old_save	*mds_prevq;
	int			mds_flags;
	struct buf		*mds_bp;
	struct buf		*mds_back;
	int			mds_frags;
	int			mds_frag_chars;
	struct md_old_save	*mds_mdsave;
};

struct md_old_struct	    {
	short	md_status;
	short	md_flag;
	short	md_partit[OLD_NMIRROR];
	int	md_total_blocks;
	short	md_rows;
	short	md_copy_size;
	int	md_current_copy;
	struct	md_old_save md_writeq;
	short	md_outstanding;
	short	md_ecount;
	short	md_err_threshold;
	short	md_pass_num;
	dev_t	md_rdev;
	struct md_old_row	{
		short	md_width;
		short	md_start_block;
		int	md_blocks;
		int	md_cum_blocks;
		dev_t	md_real_disks[MD_OLD_MAX_DISKS];
	} md_real_row[MD_OLD_MAX_ROWS];
};
